// Mapper Algorithm
// Translated from psuedocode to C++
// Original Algorithm:
//   Dr. Shahid H. Bokhari, _On the Mapping Problem_,
//       IEEE Trans. on Comp., Vol C-30, No 3, pp 207-214, March 1981
// C++ Implementation:
//   John Dunkelberg, Worcester Polytechnic Institute, March 1994

// Version #2
// This version handles the mapping of multiple problems nodes onto a
// single processing element
// it also uses a more complex fitness function based on both 
// communication costs and processor loading that is taken from
//   Talbi, E-G. and Muntean, T. "General Heuristics for the Mapping Problem"
//  

#include <iostream.h>
#include <fstream.h>
#include <string.h>

#define MAXLINE 80
#define DEBUG 0
extern "C"
{
  long random();
  int atoi(char *);
  int abs(int);
  char *strdup(const char *);
}

#include "matrix.H"

main(int argc, char **argv)
{
  int gen;                       // generation running marker
  double online,offline;         // online, offline fitness sums
  int rand_seed;

  // Arguments: problem_defintion_filename output_filename
  char *mesh_fname, *out_fname;
  mesh_fname = new char[MAXLINE];
  out_fname = new char[MAXLINE];
  
  if (argc > 3)
    {
      mesh_fname = strdup(argv[1]);
      out_fname = strdup(argv[2]);
      rand_seed = atoi(argv[3]);
    }
  else
    {
      cout << "\nMesh filename? ";
      cin >> mesh_fname;
      cout << "\nOutput filename? ";
      cin >> out_fname;
      cout << "\nRandom seed? ";
      cin >> rand_seed;
      cout << endl;
    }

  if (DEBUG)
    {
      cout << "Target Problem File: " << mesh_fname << endl;
      cout << "Output Filename:     " << out_fname << endl;
    }

  srandom(rand_seed);

  Matrix Mat(mesh_fname);
  Matrix Best;

  fstream out_fp(out_fname,ios::out);
  delete mesh_fname;   // we no longer need the mesh's filename
  delete out_fname;   // we no longer need the mesh's filename

  if (DEBUG)
    cout << "Started\n";

  /* Intialize the Best with the Mesh definition */
  Best = Mat;
  Mat.Fitness();
  Best.Fitness();
  
  if (DEBUG)
    {
      cout << "Initialized\n";
      Mat.Info();
      Best.Info();
    }

  // Bokhari's MAIN loop
  int alpha, beta;
  int best_alpha, best_beta;
  double gain, best_gain;
  int done,flag;
  done = 0;

  /* Some commonly called, unchanging values */
  int mach_size, prob_size;
  mach_size = Mat.Get_Mach_Size();
  prob_size = Mat.Get_Size();

  gen = 0;
  online = offline = 0.0;    // online and offline fitness values
  while (!(done))
    {
      if (DEBUG > 10)
	cout << "MAIN\n";
      // Bokhari's SEARCH loop
      if (DEBUG > 10)
	cout << "SEARCH \n";
      do
	{
	  flag = 0;
	  // Bokhari's AUGMENT loop : for each node do
	  best_alpha = best_beta = -1;
	  best_gain = -1.0;

	  for(alpha=0;(alpha < prob_size);alpha++)
	    {
	      // ALPHA LOOP
	      for(beta=0;(beta < prob_size);beta++)
		{
		  // BETA LOOP
		  if (DEBUG > 20)
		    cout << "AUGMENT: [" << alpha << "," << beta << "]\n";
		  // What if we set alpha's mapping to beta
		  gain = Mat.If_Set(alpha,beta);
		  if (DEBUG > 20)
		    cout << "Gain = " << gain << endl;
		  if (gain > best_gain)
		    {
		      // New best exchange found
		      best_gain = gain;
		      best_alpha = alpha;
		      best_beta = beta;
		    }
		}
	    }
	  if (best_gain >= 0.0)
	    {
	      Mat.Set(best_alpha,best_beta);
	      if (DEBUG > 10)
		{
		  cout << "Best Gain = " << best_gain;
		  cout << " Alpha " << best_alpha;
		  cout << " Beta " << best_beta << endl;
		}
	      if (best_gain > 0)
		{
		  flag = 1;
		  if (DEBUG > 10)
		    Mat.Info();
		}
	    }

	  // Output information
	  // This is not truely De Jong information, as Bokhari does not
	  // have "populations" as such.  i.e. the "population" does not
	  // follow a real trend
	  gen++;
	  online  += Mat.Fitness();
	  offline += ((Best.Fitness() > Mat.Get_Fit())
		      ? Best.Fitness() : Mat.Get_Fit());
	  out_fp <<gen <<" " <<(online/((double) gen));
	  out_fp <<" " <<(offline/((double)gen)) <<endl;

	  // END AUGMENT
	  if (DEBUG > 10)
	    cout << "END AUGMENT\n";
	} while (flag); // END SEARCH

      if (DEBUG > 10)
	cout << "END SEARCH\n";
      if ((Mat.Fitness() < Best.Fitness()) ||
	  (Mat.Fitness() == Mat.Get_Max_Edges()))
	done = 1;    // right, that's it, let's quit this program
      else
	{
	  // Bokhari's JUMP
	  if (DEBUG > 10)
	    cout << "JUMP\n";
	  Best = Mat;
	  // randomly interchange fem_size pairs of nodes of Mat
	  // in an attempt to find a better locale
	  Mat.Randomize();
	  if (DEBUG > 10)
	    Mat.Info();
	}
    } // END MAIN (done)

  if (DEBUG > 10)
    cout << "END MAIN\n";
  Best.Info();
}
