// Bokhari Adjacency Matrix

class Matrix
{
  int size;       // number of nodes in the mapping problem
  int connect;    // maximum connectivity of any given node

  int max_edges;  // number of edges in the problem

  int mach_side;  // length of a side of the target machine
  int mach_size;  // total size of the target machine (# PEs)

  double fitness;    // connectivity fitness (modified manhattan)

  // What problem elements are mapped to what machine elements
  int *map;

  // What are the connections of the problem
  int *edge;

public:
// Construction and Destruction
  Matrix(char *);
  ~Matrix()
    { delete map; };
  Initialize(fstream);
  Matrix& operator=(const Matrix&);

// Internal Actions Functions
  int Randomize();               // randomly swap around mach_side nodes
  double If_Swap(int,int);       // return the gain if we did this swap
  int Swap(int,int);             // do this swap
  double If_Set(int,int);        // return the gain if we did this set
  int Set(int,int);              // do this set
  double Fitness();              // get a fitness number

// General Information Functions
  int Get_Map(int) const;        // get a particular mapping
  int Get_Edge(int,int) const;   // get a particular edge of the problem
  inline double Get_Fit(void) const
    { return fitness; }     // return the fitness
  inline int Get_Size(void) const
    { return size; }        // get the size of the mesh (nodes)
  inline int Get_Connect(void) const
    { return connect; }     // get the connectivity of the mesh
  inline int Get_Mach_Side(void) const
    { return mach_side; }   // get the machine side length
  inline int Get_Mach_Size(void) const
    { return mach_size; }   // get the machine size
  inline int Get_Max_Edges(void) const
    { return max_edges; }   // get the number of edges
  void Info(void);
};

// Constructor calls the Problem constructor with the fstream
Matrix::Matrix(char *fname=(char *)NULL)
{
  size = connect = mach_side = mach_size = max_edges = -1;
  fitness = -1.0;
  map = (int *) NULL;
  edge = (int *) NULL;

  if (DEBUG > 30)
    cout << "Matrix::Matrix\n";

  if (fname != (char *) NULL)
    {
      fstream fp(fname,ios::in);
      if (fp.good())
	Initialize(fp);
      else
	{
	  cout << "Matrix::Matrix Def'n File cannot be opened \n";
	  exit(1);
	}

      // Now that we've loaded the file, finish construction
      // Make a default mapping of the problem
      // Which just layers them onto the machine, left to right, up to down
      int x,y;
      for(x=0;(x<size);x++)
	map[x] = x % mach_size;

      // Figure the number of edges in the problem
      max_edges = 0;
      for(x=0;(x<size);x++)
	for(y=0;(y<connect);y++)
	  if (Get_Edge(x,y) != -1)
	    max_edges++;
      max_edges = max_edges /2;
      return;
    }
  else
    {
      if (DEBUG)
	cout << "Matrix::Matrix Def'n File is NULL \n";
    }
  
}

// Initialization of problem via file desription
Matrix::Initialize(fstream fp)
{
  int node,side;
  char *com_buf,c;

  com_buf = new char[MAXLINE];

  // problem initializer
  while (fp.peek() == '#')
    {
      fp.get(com_buf,MAXLINE,'\n');
      if (DEBUG > 50)
	cout << "Disregard: " << com_buf << endl;
      fp.get(c);
    }
  fp >> mach_side;
  fp.get(c);
  mach_size = mach_side * mach_side;
  while (fp.peek() == '#')
    {
      fp.get(com_buf,MAXLINE,'\n');
      if (DEBUG > 50)
	cout << "Disregard: " << com_buf << endl;
      fp.get(c);
    }
  fp >> size;
  fp.get(c);
  while (fp.peek() == '#')
    {
      fp.get(com_buf,MAXLINE,'\n');
      if (DEBUG > 50)
	cout << "Disregard: " << com_buf << endl;
      fp.get(c);
    }
  fp >> connect;
  fp.get(c);

  if (DEBUG > 20)
    cout << "ms:"<<mach_side<<" s:"<<size<<" c:"<<connect <<endl;

  mach_size = mach_side*mach_side;
  // Now that we know how big the problem is, allocate space for it
  edge = new int[size*connect];
  map = new int[size];

  for(node=0;(node<size);node++)
    {
      for(side=0;(side<connect);side++)
	{
	  while (fp.peek() == '#')
	    {
	      fp.get(com_buf,MAXLINE,'\n');
	      if (DEBUG > 50)
		cout << "Disregard: " << com_buf << endl;
	      fp.get(c);
	    }
	  fp >> edge[(node*connect)+side];
	  fp.get(c);
	}
    }
  
  // Stop that memory leak
  delete com_buf;
}

// Copy operator
Matrix& Matrix::operator=(const Matrix& from)
{
  size = from.Get_Size();
  connect = from.Get_Connect();
  mach_side = from.Get_Mach_Side();
  mach_size = from.Get_Mach_Size();
  max_edges = from.Get_Max_Edges();
  fitness = from.Get_Fit();

  // Reassign Memory for the target
  if (map != (int *) NULL)
    delete map;
  if (edge != (int *) NULL)
    delete edge;
  map = new int[size];
  edge = new int[size*connect];

  int x,y;

  // Transfer the mapping for the machine
  for(x=0;(x<size);x++)
    map[x] = from.Get_Map(x);

  // Transfer the problem description
  for(x=0;(x<size);x++)
      for(y=0;(y<connect);y++)
	edge[(x*connect)+y] = from.Get_Edge(x,y);
  return *this;
}

int Matrix::Randomize()
{
  int count,a,b;
  for(count=0;(count<mach_side);count++)
    {
      a = ((int) random()) % size;
      b = ((int) random()) % size;
      Swap(a,b);
    }
  Fitness();
  return(count);
}

// If we were to swap alpha and beta, then return the gain in fitness
double Matrix::If_Swap(int alpha,int beta)
{
  double gain;

  static Matrix temp;
  temp = *this;

  if (DEBUG > 50)
    {
      cout << "If_Swap Temp:\n";
      temp.Info();
    }

  temp.Swap(alpha,beta);

  gain = temp.Get_Fit() - Get_Fit();
  return(gain);
}

// actually do the swap of alpha and beta in the mapping
int Matrix::Swap(int alpha,int beta)
{
  int temp;
  temp = map[beta];
  map[beta] = map[alpha];
  map[alpha] = temp;
  Fitness();

  if (DEBUG>50)
    {
      cout << "After the Swap\n";
      Info();
    }
  return(1);
}

// If we were to set alpha's mapping to beta, then return the gain in fitness
double Matrix::If_Set(int alpha,int beta)
{
  double gain;

  static Matrix temp;
  temp = *this;

  if (DEBUG > 50)
    {
      cout << "If_Set Temp:\n";
      temp.Info();
    }

  temp.Set(alpha,beta);

  gain = temp.Get_Fit() - Get_Fit();
  return(gain);
}

// actually do the set of alpha's mapping to beta
int Matrix::Set(int alpha,int beta)
{
  int temp;
  map[alpha] = beta;
  Fitness();

  if (DEBUG>50)
    {
      cout << "After the Set\n";
      Info();
    }
  return(1);
}

// Fitness function
// determines the fitness of a particular mapping
// this uses both a loading and a communication system given equal weight
double Matrix::Fitness()
{
  int cur_node,cur_edge;
  int node,target;
  double maxfit;
  double comm_fit = 0.0;
  double load_fit = 0.0;

  // Communication Fitness Section
  int x1,x2,y1,y2,dx,dy,temp;
  for(cur_node=0;(cur_node < size);cur_node++)
    for(cur_edge=0,node=Get_Map(cur_node);(cur_edge < connect);cur_edge++)
      {
	if (Get_Edge(cur_node,cur_edge) != -1)
	  {
	    target = Get_Map(Get_Edge(cur_node,cur_edge));
	    
	    // Modified Manhattan Distance
	    x1 = map[node]%mach_side;
	    x2 = map[target]%mach_side;
	    dx = abs(x1 - x2);
	    temp = ((x1>x2) ? x2+abs(mach_side-x1) :
		    x1+abs(mach_side-x2));
	    dx = ((dx>temp) ? temp : dx);
	    
	    y1 = map[node]/(mach_side);
	    y2 = map[target]/(mach_side);
	    dy = abs(y1 - y2);
	    temp = ((y1>y2) ? y2 + abs((mach_side)-y1) :
		    y1 + abs((mach_side)-y2));
	    dy = ((dy>temp) ? temp : dy);
	    comm_fit += (double) ((dx>dy) ? dy : dx) + abs(dx-dy);
	  }
      }
  comm_fit = comm_fit / 2.0;
  maxfit = (double) (max_edges * (mach_side/2));
  comm_fit = 1.0 - (comm_fit/maxfit);

  // Loading Fitness Section
  double avg_load, run_total;
  double tmp_load;
  int *map_total;

  // Compile the load on each processor
  map_total = new int[mach_size];
  // Intialize temporary totalling array
  for(target=0;(target < mach_size);target++)
    map_total[target] = 0;
  // increment the appropos map_total for each mapping of a node
  for(node =0;(node < size);node++)
    map_total[Get_Map(node)]++;

  // figure the average load on a given PE
  avg_load = (double) size / (double) mach_size;

  // figure the variance of each PE's load, keep a running total
  for(target=0, run_total=0.0;(target < mach_size);target++)
    {
      tmp_load = (double) map_total[target];
      run_total += (tmp_load*tmp_load) - (avg_load*avg_load);
    }
  load_fit = run_total / (double) mach_size;
  // Figure worst case
  tmp_load = (double) (size*size) - ((double) mach_size - 1.0) * avg_load;
  load_fit = (tmp_load - load_fit)/tmp_load;
  delete map_total;   // reclaim memory of map loading array

  // fitness is equal weighting of load fitness and communication fitness
  fitness = (comm_fit + load_fit) / 2.0;

//  cout <<fitness <<"= comm:" <<comm_fit <<" & load:" <<load_fit <<endl;

  return(fitness);
}

// Get a particular mapping... a bit too large for an inline
int Matrix::Get_Map(int x) const
{
  if ((x >=0) && (x < size))
    return(map[x]);
  else
    return(-1);        // no such machine node to map
}

// Get a particular edge of the problem... a bit too large for an inline
int Matrix::Get_Edge(int x,int y) const
{
  if ((x >= 0) && (x < size) &&
      (y >= 0) && (y < connect))
    {
      return(edge[(x*connect) + y]);
    }
  else
    return(-1);         // no such edge
}

void Matrix::Info()
{
  int x,y;

  cout << "Fitness: "<< fitness<< " Max Fitness (edges): "<< max_edges<< endl;
  cout << "Size: " << size << " Connect: " << connect << endl;
  if (map == (int *) NULL)
    cout << "Mapping: NULL\n";
  else
    {
      for(x=0, cout << "Mapping: ";(x<size);x++)
	cout << "[" << Get_Map(x) << "] ";
      cout << endl;
    }
  if (DEBUG > 50)
    if (edge == (int *) NULL)
      cout << "Edge: NULL\n";
    else
      {
	for(x=0, cout << "Edges: ";(x<size);x++)
	  for(y=0, cout << endl;(y<connect);y++)
	    cout << "[" << Get_Edge(x,y) << "] ";
	cout << endl;
      }
}

